keypad > grid > button, .dial-button, .delete-button .rounded-button {
  border-radius: 9999px;
}

.dial-button {
  padding: 19px 42px;
}

.delete-button {
  padding: 8px 12px;
}

.phone-number-entry {
  font-size: 24px;
}

.address-entry {
  font-size: 18px;
}

.error-state-message {
  background-color: @error_color;
  border-bottom: 1px solid darker(@error_color);
  color: @white;
  padding: 6px;
}

.no-background {
  background: none;
}
